/*
 *  Teletext EPG acquisition control
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation. You find a copy of this
 *  license in the file COPYRIGHT in the root directory of this release.
 *
 *  THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL,
 *  BUT WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
 *  MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *
 *  Description:
 *
 *    This module controls the Teletext EPG grabbing process. This
 *    covers cycling through the channel table, controlling the teletext
 *    packet grabber and post-processing, monitoring for channel changes
 *    and monitoring the acquisition progress.
 *
 *  Author: Tom Zoerner
 *
 *  $Id: epgacqttx.c,v 1.2 2007/03/10 19:32:23 tom Exp tom $
 */

#define DEBUG_SWITCH DEBUG_SWITCH_EPGCTL
#define DPRINTF_OFF

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include "epgctl/mytypes.h"
#include "epgctl/debug.h"

#include "epgvbi/btdrv.h"
#include "epgvbi/ttxdecode.h"
#include "epgdb/ttxgrab.h"
#include "epgdb/epgblock.h"
#include "epgdb/epgdbif.h"
#include "epgdb/epgdbsav.h"

#include "epgui/uictrl.h"
#include "epgui/epgmain.h"
#include "epgctl/epgctxctl.h"
#include "epgctl/epgacqctl.h"
#include "epgctl/epgacqsrv.h"
#include "epgctl/epgacqttx.h"


// ---------------------------------------------------------------------------
// Declaration of module state variables
//
typedef enum
{
   TTXACQ_STATE_OFF,               // acquisition disabled
   TTXACQ_STATE_STARTUP,           // now-scan / wait for page header
   TTXACQ_STATE_GRAB,              // grab teletext (wait for timeout or 'til range done)
   TTXACQ_STATE_GRAB_PASSIVE,      // passive capture
   TTXACQ_STATE_IDLE,              // passive done or channel unknown
} TTXACQ_STATE;

typedef struct
{
   bool           srcDone;
   time_t         mtime;
   uint           freq;
} TTXACQ_SOURCE;

typedef struct
{
   TTXACQ_STATE   state;
   EPGACQ_MODE    mode;
   EPGACQ_PASSIVE passiveReason;
   EPGACQ_PHASE   cyclePhase;
   uint           ttxSrcCount;
   sint           ttxSrcIdx;
   time_t         acqStartTime;
   time_t         chanChangeTime;
   bool           allSrcDone;
   bool           srcSkipWrapped;
   sint           lastDoneSrc;
   char          *pTtxNames;
   TTXACQ_SOURCE *pSources;
} EPGACQTTX_CTL;

static EPGACQTTX_CTL  acqCtl;

#define TTX_START_TIMEOUT_NOW         3
#define TTX_START_TIMEOUT_FULL       15
#define TTX_START_TIMEOUT_IDLE       60

#define TTX_GRAB_TIMEOUT_NOW         25
#define TTX_GRAB_TIMEOUT_FULL        90
#define TTX_GRAB_TIMEOUT_PASV_NOW    30
#define TTX_GRAB_TIMEOUT_PASV_FULL  120
#define TTX_GRAB_TIMEOUT_IDLE       (30*60)

#define TTX_FORCED_PASV_INTV         20
#define TTX_SLICER_CHECK_INTV        20

// number of seconds without TTX header after which acq is considered "stalled"
#define ACQ_DESCR_TTX_STALLED_TIMEOUT  7

#ifdef USE_TTX_GRABBER
static bool EpgAcqTtx_UpdateProvider( void );

// ---------------------------------------------------------------------------
// Helper functions which looks up a channel's XML file name
//
static const char * EpgAcqTtx_GetChannelName( sint ttxSrcIdx )
{
   const char * pNames = "";
   uint idx;

   if (acqCtl.pTtxNames != NULL)
   {
      if ((ttxSrcIdx >= 0) && (ttxSrcIdx < acqCtl.ttxSrcCount))
      {
         pNames = acqCtl.pTtxNames;
         for (idx = 0; idx < ttxSrcIdx; idx++)
            while(*(pNames++) != 0)
               ;
      }
      else
         ifdebug2(ttxSrcIdx != -1, "EpgAcqTtx-GetChannelName: invalid idx %d (>=%d)", ttxSrcIdx, acqCtl.ttxSrcCount);
   }

   return pNames;
}
#endif

// ---------------------------------------------------------------------------
// Determine state of acquisition
//
void EpgAcqTtx_DescribeAcqState( EPGACQ_DESCR * pAcqState )
{
   uint idx;
   time_t now = time(NULL);

   switch (acqCtl.state)
   {
      case TTXACQ_STATE_OFF:
         pAcqState->ttxGrabState = ACQDESCR_DISABLED;
         break;

      case TTXACQ_STATE_STARTUP:
         if (now - acqCtl.chanChangeTime > ACQ_DESCR_TTX_STALLED_TIMEOUT)
            pAcqState->ttxGrabState = ACQDESCR_NO_RECEPTION;
         else if (acqCtl.cyclePhase == ACQMODE_PHASE_NOWNEXT)
            pAcqState->ttxGrabState = ACQDESCR_TTX_PG_SEQ_SCAN;
         else
            pAcqState->ttxGrabState = ACQDESCR_STARTING;
         break;

      case TTXACQ_STATE_GRAB:
      case TTXACQ_STATE_GRAB_PASSIVE:
         pAcqState->ttxGrabState = ACQDESCR_RUNNING;
         break;

      case TTXACQ_STATE_IDLE:
         pAcqState->ttxGrabState = ACQDESCR_IDLE;
         break;
   }

   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      pAcqState->ttxSrcCount = acqCtl.ttxSrcCount;
      pAcqState->ttxGrabIdx = acqCtl.ttxSrcIdx;

      pAcqState->ttxGrabDone = 0;
      for (idx = 0; idx < acqCtl.ttxSrcCount; idx++)
         if (acqCtl.pSources[idx].srcDone)
            pAcqState->ttxGrabDone += 1;
   }
}

// ---------------------------------------------------------------------------
// Return complete set of acq state and statistic values
// - used by "View acq statistics" popup window
//
bool EpgAcqTtx_GetAcqStats( EPG_TTX_GRAB_STATS * pTtxGrabStats )
{
   bool result = FALSE;

#ifdef USE_TTX_GRABBER
   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      TtxGrab_GetStatistics(&pTtxGrabStats->pkgStats);

      pTtxGrabStats->acqStartTime = acqCtl.acqStartTime;
      pTtxGrabStats->srcIdx = acqCtl.ttxSrcIdx;

      strncpy(pTtxGrabStats->srcName, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx), EPG_TTX_STATS_NAMLEN);
      pTtxGrabStats->srcName[EPG_TTX_STATS_NAMLEN - 1] = 0;

      result = TRUE;
   }
#endif
   return result;
}

#ifdef USE_TTX_GRABBER
// ---------------------------------------------------------------------------
// Start teletext decoding and initialize the EPG stream decoder
//
static bool EpgAcqTtx_TtxStart( void )
{
   bool result = FALSE;

   // TODO: use generic params if source is yet unknown (ttxSrcIdx == -1)
   if (acqCtl.ttxSrcIdx != -1)
   {
      result = TtxGrab_Start(0x300, 0x399, TRUE);  // TODO page range must be passed down by GUI
      TtxDecode_StartTtxAcq(TRUE, 0x300, 0x399);
   }
   else
   {
      result = TtxGrab_Start(0x100, 0x100, FALSE);
      TtxDecode_StartTtxAcq(TRUE, 0x100, 0x100);
   }

#ifdef USE_DAEMON
   EpgAcqServer_TriggerStats();
#endif
   return result;
}

// ---------------------------------------------------------------------------
// Stop teletext decoding and clear the EPG stream and its queue
//
static void EpgAcqTtx_TtxStop( void )
{
   TtxDecode_StopTtxAcq();
   TtxGrab_Stop();
}
#endif // USE_TTX_GRABBER

// ---------------------------------------------------------------------------
// Start the acquisition
//
bool EpgAcqTtx_Start( EPGACQ_MODE mode, EPGACQ_PHASE cyclePhase )
{
   bool result = FALSE;

#ifdef USE_TTX_GRABBER
   if (acqCtl.ttxSrcCount > 0)
   {
      dprintf2("EpgAcqTtx-Start: starting acquisition phase %d, mode %d\n", cyclePhase, mode);

      acqCtl.mode       = mode;
      acqCtl.cyclePhase = cyclePhase;
      // note ttxSrcIdx is initialized separately

      acqCtl.acqStartTime = time(NULL);
      acqCtl.chanChangeTime = acqCtl.acqStartTime;
      acqCtl.state = TTXACQ_STATE_STARTUP;

      // set input source and tuner frequency (also detect if device is busy)
      EpgAcqTtx_UpdateProvider();
      result = EpgAcqTtx_TtxStart();

      UiControlMsg_AcqEvent(ACQ_EVENT_PROV_CHANGE);
   }
#endif // USE_TTX_GRABBER

   return result;
}

// ---------------------------------------------------------------------------
// Stop the acquisition
//
void EpgAcqTtx_Stop( void )
{
#ifdef USE_TTX_GRABBER
   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      dprintf0("EpgAcqTtx-Stop: stopping acquisition\n");

      EpgAcqTtx_TtxStop();

      acqCtl.state = TTXACQ_STATE_OFF;
   }
#endif
}

// ---------------------------------------------------------------------------
// Stop acquisition for the EPG scan but keep the device open/driver loaded
//
void EpgAcqTtx_Suspend( void )
{
#ifdef USE_TTX_GRABBER
   EpgAcqTtx_Stop();
#endif
}

#ifdef USE_TTX_GRABBER
// ---------------------------------------------------------------------------
// Switch database and TV channel to the current acq provider & handle errors
// - called upon start of acquisition or configuration changes (db or acq mode)
// - called upon the start of a cycle phase for one db, or to get out of
//   forced passive mode
//
static bool EpgAcqTtx_UpdateProvider( void )
{
   bool result = FALSE;
   uint freq;
   time_t now = time(NULL);

   if ( (acqCtl.mode != ACQMODE_PASSIVE) && (acqCtl.ttxSrcIdx >= 0) )
   {
      freq = acqCtl.pSources[acqCtl.ttxSrcIdx].freq;
      if (freq != 0)
      {
         if ( (acqCtl.mode != ACQMODE_PASSIVE) &&
              ((acqCtl.passiveReason == ACQPASSIVE_NONE) || (acqCtl.passiveReason != ACQPASSIVE_NO_TUNER)) )
         {
            acqCtl.chanChangeTime = now;

            result = EpgAcqCtl_TuneProvider(TRUE, freq, TRUE, &acqCtl.passiveReason);
         }
      }
      else
      {  // no channel to be tuned onto -> set at least the input source
         acqCtl.chanChangeTime = now;

         result = EpgAcqCtl_TuneProvider(TRUE, 0, FALSE, &acqCtl.passiveReason);
      }
   }
   else
   {  // passive mode
      acqCtl.chanChangeTime = now;
   }

   acqCtl.state = TTXACQ_STATE_STARTUP;
   if (result)
      acqCtl.acqStartTime = now;
   else
      acqCtl.ttxSrcIdx = -1;

   return result;
}
#endif // USE_TTX_GRABBER

// ---------------------------------------------------------------------------
// Initialize the state machine which cycles through TTX sources
// - called upon start and when switching from Nextview to teletext
//
void EpgAcqTtx_InitCycle( bool isTtxSrc, EPGACQ_PHASE phase )
{
   uint idx;

   for (idx = 0; idx < acqCtl.ttxSrcCount; idx++)
   {
      acqCtl.pSources[idx].srcDone = FALSE;
   }
   acqCtl.ttxSrcIdx = 0;
   acqCtl.allSrcDone = FALSE;
   acqCtl.srcSkipWrapped = FALSE;
   acqCtl.lastDoneSrc = -1;
}

#ifdef USE_TTX_GRABBER
// ---------------------------------------------------------------------------
// Search source table for next unfinished source
// - skip sources which already have been processes (done set to TRUE)
// - return -1 if all sources have been done
// - sets "wrap" flag if all remaining sources have been skipped once
//
static sint EpgAcqTtx_GetNextSrc( sint ttxSrcIdx )
{
   uint skipCount;
   sint result = -1;

   skipCount = 0;
   do
   {
      if ((ttxSrcIdx >= 0) && (ttxSrcIdx + 1 < acqCtl.ttxSrcCount))
         ttxSrcIdx = ttxSrcIdx + 1;
      else
         ttxSrcIdx = 0;

      if ((ttxSrcIdx == acqCtl.lastDoneSrc) && (acqCtl.srcSkipWrapped == FALSE))
      {
         dprintf1("EpgAcqTtx-GetNextSrc: search wrapped around last idx=%d\n", acqCtl.lastDoneSrc);
         acqCtl.srcSkipWrapped = TRUE;
      }
      if (acqCtl.pSources[ttxSrcIdx].srcDone == FALSE)
      {
         result = ttxSrcIdx;
         break;
      }
      skipCount += 1;
   } while (skipCount < acqCtl.ttxSrcCount);

   return result;
}

// ---------------------------------------------------------------------------
// Try to determine current source in (forced) passive mode
// - TODO: external sources: frequency match not possible! (Must still support passive mode)
//
static bool EpgAcqTtx_DetectSource( void )
{
   uint freq;
   uint input;
   bool isTuner;
   uint idx;
   bool result = FALSE;

   if (BtDriver_QueryChannel(&freq, &input, &isTuner))
   {
      if (isTuner)
      {
         for (idx = 0; idx < acqCtl.ttxSrcCount; idx++)
            if (acqCtl.pSources[idx].freq == freq)
               break;
         if (idx < acqCtl.ttxSrcCount)
         {
            dprintf3("EpgAcqTtx-DetectSource: freq %d -> src=%d (%s)\n", freq, idx, EpgAcqTtx_GetChannelName(idx));
            acqCtl.ttxSrcIdx = idx;
            result = TRUE;
         }
         else
         {
            dprintf1("EpgAcqTtx-DetectSource: freq %d unknown\n", freq);
            acqCtl.ttxSrcIdx = -1;
         }
      }
      else
         dprintf0("EpgAcqTtx-DetectSource: source is not tuner\n");
   }
   else
      debug0("EpgAcqTtx-DetectSource: failed to query freq from driver");

   return result;
}

// ---------------------------------------------------------------------------
// Check if the requested channel is still tuned
// - TODO: external sources: frequency match not possible!
//
static bool EpgAcqTtx_CheckSourceFreq( sint ttxSrcIdx )
{
   uint freq;
   uint input;
   bool isTuner;
   bool result = FALSE;

   if ((ttxSrcIdx >= 0) && (ttxSrcIdx < acqCtl.ttxSrcCount))
   {
      if (BtDriver_QueryChannel(&freq, &input, &isTuner))
      {
         result = isTuner && (acqCtl.pSources[ttxSrcIdx].freq == freq);
         ifdebug3(!result && isTuner, "EpgAcqTtx-CheckSourceFreq: freq %d mismatch to src=%d freq %d", freq, acqCtl.pSources[ttxSrcIdx].freq, ttxSrcIdx);
      }
      else
         debug0("EpgAcqTtx-CheckSourceFreq: failed to query freq from driver");
   }
   return result;
}
#endif // USE_TTX_GRABBER

// ---------------------------------------------------------------------------
// Handle external channel changes
// - note: NOT called upon internally controlled channel changes
//
void EpgAcqTtx_ChannelChange( void )
{
#ifdef USE_TTX_GRABBER
   time_t now;

   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      if ( (acqCtl.state == TTXACQ_STATE_GRAB) ||
           (acqCtl.state == TTXACQ_STATE_GRAB_PASSIVE) )
      {
         assert(acqCtl.ttxSrcIdx != -1);

         now = time(NULL);
         if (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_PASV_NOW)
         {
            TtxGrab_PostProcess(EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx), TRUE);
         }
      }

      // restart acquisition
      acqCtl.acqStartTime = time(NULL);
      acqCtl.chanChangeTime = acqCtl.acqStartTime;
      acqCtl.state = TTXACQ_STATE_STARTUP;
      EpgAcqTtx_DetectSource();

      EpgAcqTtx_TtxStart();
      UiControlMsg_AcqEvent(ACQ_EVENT_CTL);
   }
#endif // USE_TTX_GRABBER
}

// ---------------------------------------------------------------------------
// Monitor acquisition progress
// - Now&Next mode: switch to next source if target page range is not near
// - check timeouts (no packets received; page range complete)
// - determine slicer type on all pages (i.e. also outside of page range)
//   (needs faster adjustment than nxtv. to support Now&Next scan)
// - returns TRUE when all teletext sources are done (advance the cycle phase)
//   Note this differs from the Nextview state machine: here we advance only
//   after ALL teletext sources are done, i.e. switching between teletext
//   sources is handled internally (since it's not always sequential)
//
bool EpgAcqTtx_MonitorSources( void )
{
#ifdef USE_TTX_GRABBER
   sint ttxSrcIdx;
   bool inRange;
   bool rangeDone;
   bool lock;
   time_t now = time(NULL);
   bool advance = FALSE;

   if (TtxGrab_CheckPostProcess(NULL))
   {
#ifdef USE_DAEMON
      // trigger the server to notify all connected clients
      EpgAcqServer_TriggerDbUpdate(time(NULL));
#endif
      UiControlMsg_AcqEvent(ACQ_EVENT_NEW_DB);
   }

   if (acqCtl.state == TTXACQ_STATE_STARTUP)
   {
      TtxGrab_GetPageStats(&inRange, NULL, &lock);

      if ( (acqCtl.mode != ACQMODE_PASSIVE) &&
           (acqCtl.mode != ACQMODE_EXTERNAL) &&
           (acqCtl.passiveReason == ACQPASSIVE_NONE) )
      {
         assert((acqCtl.ttxSrcIdx >= 0) && (acqCtl.ttxSrcIdx < acqCtl.ttxSrcCount));

         if (acqCtl.cyclePhase == ACQMODE_PHASE_NOWNEXT)
         {
            // NOW mode: query page sequence prediction
            if ((inRange == FALSE) && (acqCtl.srcSkipWrapped == FALSE))
            {
               dprintf2("EpgAcqTtx-MonitorSources: SKIP channel %d (%s)\n", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx));
               advance = TRUE;
            }
            else if (lock)
            {
               if (EpgAcqTtx_CheckSourceFreq(acqCtl.ttxSrcIdx))
               {
                  dprintf2("EpgAcqTtx-MonitorSources: load NOW from channel %d (%s)\n", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx));
                  acqCtl.state = TTXACQ_STATE_GRAB;
                  UiControlMsg_AcqEvent(ACQ_EVENT_CTL);
               }
               else
                  EpgAcqTtx_ChannelChange();
            }
            else if (now >= acqCtl.chanChangeTime + TTX_START_TIMEOUT_NOW)
            {
               debug2("EpgAcqTtx-MonitorSources: no reception on %d (%s) - mark DONE", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx));
               acqCtl.pSources[acqCtl.ttxSrcIdx].srcDone = TRUE;
               advance = TRUE;
            }
         }
         else
         {
            if (lock)
            {
               if (EpgAcqTtx_CheckSourceFreq(acqCtl.ttxSrcIdx))
               {
                  acqCtl.state = TTXACQ_STATE_GRAB;
                  UiControlMsg_AcqEvent(ACQ_EVENT_CTL);
               }
               else
                  EpgAcqTtx_ChannelChange();
            }
            else if (now >= acqCtl.chanChangeTime + TTX_START_TIMEOUT_FULL)
            {
               debug2("EpgAcqTtx-MonitorSources: no reception on %d (%s) - advance", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx));
               acqCtl.pSources[acqCtl.ttxSrcIdx].srcDone = TRUE;
               advance = TRUE;
            }
         }
      }
      else
      {  // passive mode
         if (lock)
         {
            if (acqCtl.ttxSrcIdx == -1)
            {  // channel yet unknown -> try to detect by frequency
               if ( EpgAcqTtx_DetectSource() )
               {  // identified -> restart acq with correct parameters
                  acqCtl.acqStartTime = time(NULL);
                  EpgAcqTtx_TtxStart();
               }
               else
               {  // unknown
                  acqCtl.state = TTXACQ_STATE_IDLE;
               }
            }
            else
            {  // channel known (i.e. we restarted after a frequency match - see above)
               // check if frequency still matches
               if (EpgAcqTtx_CheckSourceFreq(acqCtl.ttxSrcIdx))
               {
                  // match -> enable acquisition
                  dprintf2("EpgAcqTtx-MonitorSources: load passive from channel %d (%s)\n", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx));
                  acqCtl.state = TTXACQ_STATE_GRAB_PASSIVE;
                  UiControlMsg_AcqEvent(ACQ_EVENT_PROV_CHANGE);
               }
               else
               {  // frequency mismatch -> restart
                  EpgAcqTtx_UpdateProvider();
                  EpgAcqTtx_TtxStart();
               }
            }
         }
         else if (now >= acqCtl.chanChangeTime + TTX_START_TIMEOUT_IDLE)
         {
            // reset acquisition to prevent getting a huge output file
            EpgAcqTtx_TtxStart();
         }
      }
   }
   else if (acqCtl.state == TTXACQ_STATE_GRAB)
   {
      assert((acqCtl.ttxSrcIdx >= 0) && (acqCtl.ttxSrcIdx < acqCtl.ttxSrcCount));

      if (acqCtl.cyclePhase == ACQMODE_PHASE_NOWNEXT)
      {
         TtxGrab_GetPageStats(NULL, &rangeDone, NULL);
         if (rangeDone || (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_NOW))
         {
            advance = TRUE;
         }
         //TtxGrab_GetPageStats(&inRange, NULL, NULL);
         //if ((inRange == FALSE) && (pageCount <= 1))
         // -> revert decisition to stay on this channel -> advance
      }
      else
      {
         advance = (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_FULL);
      }
   }
   else if (acqCtl.state == TTXACQ_STATE_GRAB_PASSIVE)
   {
      assert((acqCtl.ttxSrcIdx >= 0) && (acqCtl.ttxSrcIdx < acqCtl.ttxSrcCount));

      if (acqCtl.cyclePhase == ACQMODE_PHASE_NOWNEXT)
         advance = (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_PASV_NOW);
      else
         advance = (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_PASV_FULL);
   }
   else if (acqCtl.state == TTXACQ_STATE_IDLE)
   {
      advance = (now >= acqCtl.chanChangeTime + TTX_GRAB_TIMEOUT_IDLE);
   }
   else
      fatal1("EpgAcqTtx-MonitorSources: invalid acq state %d", acqCtl.state);

   if (advance)
   {
      if ( (acqCtl.state == TTXACQ_STATE_GRAB) ||
           (acqCtl.state == TTXACQ_STATE_GRAB_PASSIVE) )
      {
         TtxGrab_PostProcess(EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx), TRUE);
         acqCtl.pSources[acqCtl.ttxSrcIdx].srcDone = TRUE;
         acqCtl.lastDoneSrc = acqCtl.ttxSrcIdx;
      }
      if ( (acqCtl.mode != ACQMODE_PASSIVE) &&
           (acqCtl.mode != ACQMODE_EXTERNAL) )
      {
         if (acqCtl.cyclePhase == ACQMODE_PHASE_NOWNEXT)
            ttxSrcIdx = EpgAcqTtx_GetNextSrc(acqCtl.ttxSrcIdx);
         else // note: -1 to restore normal order after passive mode
            ttxSrcIdx = EpgAcqTtx_GetNextSrc(-1);
         if (ttxSrcIdx >= 0)
         {
            dprintf4("EpgAcqTtx-MonitorSources: %d (%s) done, next channel %d (%s)\n", acqCtl.ttxSrcIdx, EpgAcqTtx_GetChannelName(acqCtl.ttxSrcIdx), ttxSrcIdx, EpgAcqTtx_GetChannelName(ttxSrcIdx));
            acqCtl.ttxSrcIdx = ttxSrcIdx;

            EpgAcqTtx_UpdateProvider();
            EpgAcqTtx_TtxStart();

            UiControlMsg_AcqEvent(ACQ_EVENT_PROV_CHANGE);
         }
         else
         {
            acqCtl.allSrcDone = TRUE;
            // fail-safe: enter "idle" state until top-level scheduler resets us
            acqCtl.chanChangeTime = time(NULL);
            acqCtl.state = TTXACQ_STATE_IDLE;
            EpgAcqTtx_TtxStart();
         }
      }
      else
      {  // passive capture done -> disable capturing
         acqCtl.chanChangeTime = time(NULL);
         acqCtl.state = TTXACQ_STATE_IDLE;
         acqCtl.ttxSrcIdx = -1;
         EpgAcqTtx_TtxStart();
         UiControlMsg_AcqEvent(ACQ_EVENT_PROV_CHANGE);
      }
   }
   else // try to get out of forced passive mode periodically
   if ( (acqCtl.passiveReason != ACQPASSIVE_NONE) &&
        (acqCtl.mode != ACQMODE_PASSIVE) &&
        (acqCtl.mode != ACQMODE_EXTERNAL) &&
        ((acqCtl.ttxSrcIdx == -1) || (acqCtl.pSources[acqCtl.ttxSrcIdx].srcDone)) &&
        (now >= acqCtl.chanChangeTime + TTX_FORCED_PASV_INTV) )
   {
      dprintf2("EpgAcqTtx-MonitorSources: try to end forced passive (current src:%d done:%d)\n", acqCtl.ttxSrcIdx, ((acqCtl.ttxSrcIdx == -1) ? 0:acqCtl.pSources[acqCtl.ttxSrcIdx].srcDone));
      acqCtl.ttxSrcIdx = EpgAcqTtx_GetNextSrc(-1);
      EpgAcqTtx_UpdateProvider();
      EpgAcqTtx_TtxStart();
      UiControlMsg_AcqEvent(ACQ_EVENT_PROV_CHANGE);
   }
#endif // USE_TTX_GRABBER

   return acqCtl.allSrcDone;
}

// ---------------------------------------------------------------------------
// Process TTX packets in the VBI ringbuffer
// - called about once a second (even if no packets were captured)
// - checks for unsolicited channel changes (by comparing teletext headers)
//
bool EpgAcqTtx_ProcessPackets( bool * pCheckSlicer )
{
#ifdef USE_TTX_GRABBER
   time_t now;

   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      if (TtxGrab_ProcessPackets() == FALSE)
      {
         dprintf0("EpgAcqTtx-ProcessPackets: uncontrolled channel change detected\n");
         EpgAcqTtx_ChannelChange();
      }

      // check if the current slicer type is adequate
      if (pCheckSlicer != NULL)
      {
         now = time(NULL);
         // FIXME: should be less for NOW mode
         if ( (now >= acqCtl.chanChangeTime + TTX_SLICER_CHECK_INTV) &&
              (TtxGrab_CheckSlicerQuality() == FALSE) )
         {
            *pCheckSlicer = FALSE;
         }
      }
   }
#endif // USE_TTX_GRABBER

   return FALSE;
}

// ---------------------------------------------------------------------------
// Copy parameters
// - can be already called before acquisition starts, i.e. params remain valid
//   across acquisition stop and re-enable
//
void EpgAcqTtx_SetParams( uint ttxSrcCount, const char * pTtxNames, const uint * pTtxFreqs )
{
#ifdef USE_TTX_GRABBER
   uint idx;
   const char * pNames;

   if (acqCtl.pTtxNames != NULL)
   {
      xfree(acqCtl.pTtxNames);
      acqCtl.pTtxNames = NULL;
   }
   if (acqCtl.pSources != NULL)
   {
      xfree(acqCtl.pSources);
      acqCtl.pSources = NULL;
   }
   dprintf2("EpgAcqTtx-SetParams: %d TTX sources (was:%d)\n", ttxSrcCount, acqCtl.ttxSrcCount);

   acqCtl.ttxSrcCount = ttxSrcCount;
   if (ttxSrcCount != 0)
   {
      // determine total length of name table
      pNames = pTtxNames;
      for (idx = 0; idx < ttxSrcCount; idx++)
         while(*(pNames++) != 0)
            ;
      // copy names
      acqCtl.pTtxNames = xmalloc(pNames - pTtxNames);
      memcpy(acqCtl.pTtxNames, pTtxNames, (pNames - pTtxNames));

      // copy frequency table
      acqCtl.pSources = xmalloc(ttxSrcCount * sizeof(*acqCtl.pSources));
      for (idx = 0; idx < ttxSrcCount; idx++)
      {
         acqCtl.pSources[idx].freq = pTtxFreqs[idx];
         acqCtl.pSources[idx].mtime = 0; //TODO
      }
   }

   // reset state if incompatible with new params
   if (acqCtl.state != TTXACQ_STATE_OFF)
   {
      if (acqCtl.ttxSrcIdx >= acqCtl.ttxSrcCount)
         acqCtl.ttxSrcIdx = -1;
      if (acqCtl.ttxSrcCount == 0)
         acqCtl.mode = ACQMODE_PASSIVE;
   }
#endif // USE_TTX_GRABBER
}

// ---------------------------------------------------------------------------
// Check if parameters have changed
// - used to check if acq must be reset for parameters changes while acq is running
// - returns FALSE upon changes
//
bool EpgAcqTtx_CompareParams( uint ttxSrcCount, const char * pTtxNames, const uint * pTtxFreqs )
{
#ifdef USE_TTX_GRABBER
   const char * pNames;
   uint idx;
   bool change;

   if (ttxSrcCount == acqCtl.ttxSrcCount)
   {
      if (ttxSrcCount > 0)
      {
         assert((pTtxNames != NULL) && (pTtxFreqs != NULL));
         assert((acqCtl.pTtxNames != NULL) && (acqCtl.pSources != NULL));

         // compare frequencies
         change = FALSE;
         for (idx = 0; (idx < ttxSrcCount) && !change; idx++)
            if (pTtxFreqs[idx] != acqCtl.pSources[idx].freq)
               change = TRUE;

         if (change == FALSE)
         {
            // determine total length of name table
            pNames = pTtxNames;
            for (idx = 0; idx < ttxSrcCount; idx++)
               while(*(pNames++) != 0)
                  ;
            // compare names with memcmp instead of strcmp because of 0-separators between names
            change = (memcmp(acqCtl.pTtxNames, pTtxNames, (pNames - pTtxNames)) != 0);
         }
      }
      else
      {  // acquisition is disabled and will stay disabled
         change = FALSE;
      }
   }
   else
      change = TRUE;

   return !change;
#else // USE_TTX_GRABBER
   return TRUE;
#endif
}

// ---------------------------------------------------------------------------
// Stop acquisition and free allocated resources
//
void EpgAcqTtx_Destroy( void )
{
#ifdef USE_TTX_GRABBER
   EpgAcqTtx_Stop();
   TtxGrab_Exit();

   if (acqCtl.pTtxNames != NULL)
   {
      xfree(acqCtl.pTtxNames);
      acqCtl.pTtxNames = NULL;
   }
   if (acqCtl.pSources != NULL)
   {
      xfree(acqCtl.pSources);
      acqCtl.pSources = NULL;
   }
#endif // USE_TTX_GRABBER
}

// ---------------------------------------------------------------------------
// Initialize the module
//
void EpgAcqTtx_Init( void )
{
   memset(&acqCtl, 0, sizeof(acqCtl));
   acqCtl.state     = TTXACQ_STATE_OFF;

#ifdef USE_TTX_GRABBER
   TtxGrab_Init();
#endif
}

